@@ -41,89 +41,88 @@ fn pyo3_ffi_minimum_python_minor_version(major_version: u64, minor_version: u64)
 #[serde(default, rename_all = "kebab-case")]
 pub struct CargoOptions {
     /// Do not print cargo log messages
-    #[clap(short = 'q', long)]
+    #[arg(short = 'q', long)]
     pub quiet: bool,
 
     /// Number of parallel jobs, defaults to # of CPUs
-    #[clap(short = 'j', long, value_name = "N")]
+    #[arg(short = 'j', long, value_name = "N")]
     pub jobs: Option<usize>,
 
     /// Build artifacts with the specified Cargo profile
-    #[clap(long, value_name = "PROFILE-NAME")]
+    #[arg(long, value_name = "PROFILE-NAME")]
     pub profile: Option<String>,
 
     /// Space or comma separated list of features to activate
-    #[clap(short = 'F', long, multiple_occurrences = true)]
+    #[arg(short = 'F', long, action = clap::ArgAction::Append)]
     pub features: Vec<String>,
 
     /// Activate all available features
-    #[clap(long)]
+    #[arg(long)]
     pub all_features: bool,
 
     /// Do not activate the `default` feature
-    #[clap(long)]
+    #[arg(long)]
     pub no_default_features: bool,
 
     /// Build for the target triple
-    #[clap(long, name = "TRIPLE", env = "CARGO_BUILD_TARGET")]
+    #[arg(long, value_name = "TRIPLE", env = "CARGO_BUILD_TARGET")]
     pub target: Option<String>,
 
     /// Directory for all generated artifacts
-    #[clap(long, value_name = "DIRECTORY", parse(from_os_str))]
+    #[arg(long, value_name = "DIRECTORY")]
     pub target_dir: Option<PathBuf>,
 
     /// Path to Cargo.toml
-    #[clap(short = 'm', long, value_name = "PATH", parse(from_os_str))]
+    #[arg(short = 'm', long, value_name = "PATH")]
     pub manifest_path: Option<PathBuf>,
 
     /// Ignore `rust-version` specification in packages
-    #[clap(long)]
+    #[arg(long)]
     pub ignore_rust_version: bool,
 
     /// Use verbose output (-vv very verbose/build.rs output)
-    #[clap(short = 'v', long, parse(from_occurrences), max_occurrences = 2)]
-    pub verbose: usize,
+    #[arg(short = 'v', long, action = clap::ArgAction::Count)]
+    pub verbose: u8,
 
     /// Coloring: auto, always, never
-    #[clap(long, value_name = "WHEN")]
+    #[arg(long, value_name = "WHEN")]
     pub color: Option<String>,
 
     /// Require Cargo.lock and cache are up to date
-    #[clap(long)]
+    #[arg(long)]
     pub frozen: bool,
 
     /// Require Cargo.lock is up to date
-    #[clap(long)]
+    #[arg(long)]
     pub locked: bool,
 
     /// Run without accessing the network
-    #[clap(long)]
+    #[arg(long)]
     pub offline: bool,
 
     /// Override a configuration value (unstable)
-    #[clap(long, value_name = "KEY=VALUE", multiple_occurrences = true)]
+    #[arg(long, value_name = "KEY=VALUE", action = clap::ArgAction::Append)]
     pub config: Vec<String>,
 
     /// Unstable (nightly-only) flags to Cargo, see 'cargo -Z help' for details
-    #[clap(short = 'Z', value_name = "FLAG", multiple_occurrences = true)]
+    #[arg(short = 'Z', value_name = "FLAG", action = clap::ArgAction::Append)]
     pub unstable_flags: Vec<String>,
 
     /// Timing output formats (unstable) (comma separated): html, json
-    #[clap(
+    #[arg(
         long,
         value_name = "FMTS",
-        min_values = 0,
         value_delimiter = ',',
         require_equals = true
     )]
     pub timings: Option<Vec<String>>,
 
     /// Outputs a future incompatibility report at the end of the build (unstable)
-    #[clap(long)]
+    #[arg(long)]
     pub future_incompat_report: bool,
 
     /// Rustc flags
-    #[clap(takes_value = true, multiple_values = true)]
+    #[arg(num_args = 0.., trailing_var_arg = true)]
     pub args: Vec<String>,
 }
 
@@ -144,36 +143,35 @@ pub struct BuildOptions {
     /// The default is the lowest compatible `manylinux` tag, or plain `linux` if nothing matched
     ///
     /// This option is ignored on all non-linux platforms
-    #[clap(
-        name = "compatibility",
+    #[arg(
+        id = "compatibility",
         long = "compatibility",
         alias = "manylinux",
-        parse(try_from_str),
-        multiple_values = true,
-        multiple_occurrences = true
+        num_args = 0..,
+        action = clap::ArgAction::Append
     )]
     pub platform_tag: Vec<PlatformTag>,
 
     /// The python versions to build wheels for, given as the names of the
     /// interpreters.
-    #[clap(short, long, multiple_values = true, multiple_occurrences = true)]
+    #[arg(short, long, num_args = 0.., action = clap::ArgAction::Append)]
     pub interpreter: Vec<PathBuf>,
 
     /// Find interpreters from the host machine
-    #[clap(short = 'f', long, conflicts_with = "interpreter")]
+    #[arg(short = 'f', long, conflicts_with = "interpreter")]
     pub find_interpreter: bool,
 
-    /// Which kind of bindings to use. Possible values are pyo3, rust-cpython, cffi and bin
-    #[clap(short, long)]
+    /// Which kind of bindings to use.
+    #[arg(short, long, value_parser = ["pyo3", "pyo3-ffi", "rust-cpython", "cffi", "uniffi", "bin"])]
     pub bindings: Option<String>,
 
     /// The directory to store the built wheels in. Defaults to a new "wheels"
     /// directory in the project's target directory
-    #[clap(short, long, parse(from_os_str))]
+    #[arg(short, long)]
     pub out: Option<PathBuf>,
 
     /// Don't check for manylinux compliance
-    #[clap(long = "skip-auditwheel")]
+    #[arg(long = "skip-auditwheel")]
     pub skip_auditwheel: bool,
 
     /// For manylinux targets, use zig to ensure compliance for the chosen manylinux version
@@ -181,16 +179,16 @@ pub struct BuildOptions {
     /// Default to manylinux2014/manylinux_2_17 if you do not specify an `--compatibility`
     ///
     /// Make sure you installed zig with `pip install maturin[zig]`
-    #[clap(long)]
+    #[arg(long)]
     pub zig: bool,
 
     /// Control whether to build universal2 wheel for macOS or not.
     /// Only applies to macOS targets, do nothing otherwise.
-    #[clap(long)]
+    #[arg(long)]
     pub universal2: bool,
 
     /// Cargo build options
-    #[clap(flatten)]
+    #[command(flatten)]
     pub cargo: CargoOptions,
 }
 
@@ -272,11 +270,10 @@ impl BuildOptions {
                         let ext_suffix = sysconfig_data
                             .get("EXT_SUFFIX")
                             .context("syconfig didn't define an `EXT_SUFFIX` ಠ_ಠ")?;
-                        let abi_tag = sysconfig_data
-                            .get("SOABI")
-                            .and_then(|abi| abi.split('-').nth(1).map(ToString::to_string));
-                        let interpreter_kind = sysconfig_data
-                            .get("SOABI")
+                        let soabi = sysconfig_data.get("SOABI");
+                        let abi_tag =
+                            soabi.and_then(|abi| abi.split('-').nth(1).map(ToString::to_string));
+                        let interpreter_kind = soabi
                             .and_then(|tag| {
                                 if tag.starts_with("pypy") {
                                     Some(InterpreterKind::PyPy)
@@ -300,6 +297,8 @@ impl BuildOptions {
                             executable: PathBuf::new(),
                             platform: None,
                             runnable: false,
+                            implmentation_name: interpreter_kind.to_string().to_ascii_lowercase(),
+                            soabi: soabi.cloned(),
                         });
                     } else {
                         if interpreter.is_empty() && !self.find_interpreter {
@@ -348,7 +347,7 @@ impl BuildOptions {
                 println!("🐍 Using {} to generate the cffi bindings", interpreter);
                 Ok(vec![interpreter])
             }
-            BridgeModel::Bin(None) => Ok(vec![]),
+            BridgeModel::Bin(None) | BridgeModel::UniFfi => Ok(vec![]),
             BridgeModel::BindingsAbi3(major, minor) => {
                 if target.is_windows() {
                     // Ideally, we wouldn't want to use any python interpreter without abi3 at all.
@@ -378,6 +377,8 @@ impl BuildOptions {
                             executable: PathBuf::new(),
                             platform: None,
                             runnable: false,
+                            implmentation_name: "cpython".to_string(),
+                            soabi: None,
                         }])
                     } else if let Some(interp) = interpreters.get(0) {
                         println!("🐍 Using {} to generate to link bindings (With abi3, an interpreter is only required on windows)", interp);
@@ -398,6 +399,8 @@ impl BuildOptions {
                             executable: PathBuf::new(),
                             platform: None,
                             runnable: false,
+                            implmentation_name: "cpython".to_string(),
+                            soabi: None,
                         }])
                     } else {
                         bail!("Failed to find a python interpreter");
@@ -440,6 +443,8 @@ impl BuildOptions {
                             executable: PathBuf::new(),
                             platform: None,
                             runnable: false,
+                            implmentation_name: "cpython".to_string(),
+                            soabi: None,
                         }])
                     } else if target.cross_compiling() {
                         let mut interps = Vec::with_capacity(found_interpreters.len());
@@ -566,7 +571,14 @@ impl BuildOptions {
         } else {
             // User given list of interpreters
             let interpreter = if self.interpreter.is_empty() && !target.cross_compiling() {
-                vec![PathBuf::from("python3")]
+                if cfg!(test) {
+                    match env::var_os("MATURIN_TEST_PYTHON") {
+                        Some(python) => vec![python.into()],
+                        None => vec![PathBuf::from("python3")],
+                    }
+                } else {
+                    vec![PathBuf::from("python3")]
+                }
             } else {
                 self.interpreter.clone()
             };
@@ -628,45 +640,7 @@ impl BuildOptions {
             }
         }
 
-        match bridge {
-            BridgeModel::Bin(None) => {
-                // Only support two different kind of platform tags when compiling to musl target without any binding crates
-                if platform_tags.iter().any(|tag| tag.is_musllinux()) && !target.is_musl_target() {
-                    bail!(
-                        "Cannot mix musllinux and manylinux platform tags when compiling to {}",
-                        target.target_triple()
-                    );
-                }
-
-                #[allow(clippy::comparison_chain)]
-                if platform_tags.len() > 2 {
-                    bail!(
-                        "Expected only one or two platform tags but found {}",
-                        platform_tags.len()
-                    );
-                } else if platform_tags.len() == 2 {
-                    // The two platform tags can't be the same kind
-                    let tag_types = platform_tags
-                        .iter()
-                        .map(|tag| tag.is_musllinux())
-                        .collect::<HashSet<_>>();
-                    if tag_types.len() == 1 {
-                        bail!(
-                            "Expected only one platform tag but found {}",
-                            platform_tags.len()
-                        );
-                    }
-                }
-            }
-            _ => {
-                if platform_tags.len() > 1 {
-                    bail!(
-                        "Expected only one platform tag but found {}",
-                        platform_tags.len()
-                    );
-                }
-            }
-        }
+        validate_bridge_type(&bridge, &target, &platform_tags)?;
 
         // linux tag can not be mixed with manylinux and musllinux tags
         if platform_tags.len() > 1 && platform_tags.iter().any(|tag| !tag.is_portable()) {
@@ -713,6 +687,54 @@ impl BuildOptions {
     }
 }
 
+/// Checks for bridge/platform type edge cases
+fn validate_bridge_type(
+    bridge: &BridgeModel,
+    target: &Target,
+    platform_tags: &[PlatformTag],
+) -> Result<()> {
+    match bridge {
+        BridgeModel::Bin(None) => {
+            // Only support two different kind of platform tags when compiling to musl target without any binding crates
+            if platform_tags.iter().any(|tag| tag.is_musllinux()) && !target.is_musl_target() {
+                bail!(
+                    "Cannot mix musllinux and manylinux platform tags when compiling to {}",
+                    target.target_triple()
+                );
+            }
+
+            #[allow(clippy::comparison_chain)]
+            if platform_tags.len() > 2 {
+                bail!(
+                    "Expected only one or two platform tags but found {}",
+                    platform_tags.len()
+                );
+            } else if platform_tags.len() == 2 {
+                // The two platform tags can't be the same kind
+                let tag_types = platform_tags
+                    .iter()
+                    .map(|tag| tag.is_musllinux())
+                    .collect::<HashSet<_>>();
+                if tag_types.len() == 1 {
+                    bail!(
+                        "Expected only one platform tag but found {}",
+                        platform_tags.len()
+                    );
+                }
+            }
+        }
+        _ => {
+            if platform_tags.len() > 1 {
+                bail!(
+                    "Expected only one platform tag but found {}",
+                    platform_tags.len()
+                );
+            }
+        }
+    }
+    Ok(())
+}
+
 /// Uses very simple PEP 440 subset parsing to determine the
 /// minimum supported python minor version for interpreter search
 fn get_min_python_minor(metadata21: &Metadata21) -> Option<usize> {
@@ -826,6 +848,8 @@ fn find_bindings(
         Some(("pyo3-ffi".to_string(), minor))
     } else if deps.contains_key("cpython") {
         Some(("rust-cpython".to_string(), MINIMUM_PYTHON_MINOR))
+    } else if deps.contains_key("uniffi") {
+        Some(("uniffi".to_string(), MINIMUM_PYTHON_MINOR))
     } else {
         None
     }
@@ -848,7 +872,7 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
         .iter()
         .filter_map(|pkg| {
             let name = &pkg.name;
-            if name == "pyo3" || name == "pyo3-ffi" || name == "cpython" {
+            if name == "pyo3" || name == "pyo3-ffi" || name == "cpython" || name == "uniffi" {
                 Some((name.as_ref(), pkg))
             } else {
                 None
@@ -873,8 +897,13 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
     let bridge = if let Some(bindings) = bridge {
         if bindings == "cffi" {
             BridgeModel::Cffi
+        } else if bindings == "uniffi" {
+            BridgeModel::UniFfi
         } else if bindings == "bin" {
-            BridgeModel::Bin(find_bindings(&deps, &packages))
+            // uniffi bindings don't support bin
+            let bindings =
+                find_bindings(&deps, &packages).filter(|(bindings, _)| bindings != "uniffi");
+            BridgeModel::Bin(bindings)
         } else {
             if !deps.contains_key(bindings) {
                 bail!(
@@ -887,7 +916,14 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
         }
     } else if let Some((bindings, minor)) = find_bindings(&deps, &packages) {
         if !targets.contains(&"cdylib") && targets.contains(&"bin") {
-            BridgeModel::Bin(Some((bindings, minor)))
+            if bindings == "uniffi" {
+                // uniffi bindings don't support bin
+                BridgeModel::Bin(None)
+            } else {
+                BridgeModel::Bin(Some((bindings, minor)))
+            }
+        } else if bindings == "uniffi" {
+            BridgeModel::UniFfi
         } else {
             BridgeModel::Bindings(bindings, minor)
         }
@@ -908,7 +944,7 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
             let pyo3_node = deps[lib];
             if !pyo3_node.features.contains(&"extension-module".to_string()) {
                 let version = cargo_metadata[&pyo3_node.id].version.to_string();
-                println!(
+                eprintln!(
                     "⚠️  Warning: You're building a library without activating {}'s \
                      `extension-module` feature. \
                      See https://pyo3.rs/v{}/building_and_distribution.html#linking",
@@ -1351,7 +1387,7 @@ mod test {
             .exec()
             .unwrap();
         let metadata21 =
-            Metadata21::from_cargo_toml(&cargo_toml, &"test-crates/pyo3-pure", &cargo_metadata)
+            Metadata21::from_cargo_toml(&cargo_toml, "test-crates/pyo3-pure", &cargo_metadata)
                 .unwrap();
         assert_eq!(get_min_python_minor(&metadata21), None);
     }
