@@ -10,13 +10,16 @@ use std::io::{BufReader, Read};
 use std::path::{Path, PathBuf};
 use std::process::Stdio;
 use std::str;
+use tracing::{debug, trace};
 
 /// The first version of pyo3 that supports building Windows abi3 wheel
 /// without `PYO3_NO_PYTHON` environment variable
 const PYO3_ABI3_NO_PYTHON_VERSION: (u64, u64, u64) = (0, 16, 4);
 
 /// crate types excluding `bin`, `cdylib` and `proc-macro`
-const LIB_CRATE_TYPES: [&str; 4] = ["lib", "dylib", "rlib", "staticlib"];
+pub(crate) const LIB_CRATE_TYPES: [&str; 4] = ["lib", "dylib", "rlib", "staticlib"];
+
+pub(crate) type CompileTarget<'a> = (&'a cargo_metadata::Target, &'a BridgeModel);
 
 /// A cargo build artifact
 #[derive(Debug, Clone)]
@@ -31,71 +34,46 @@ pub struct BuildArtifact {
 /// Builds the rust crate into a native module (i.e. an .so or .dll) for a
 /// specific python version. Returns a mapping from crate type (e.g. cdylib)
 /// to artifact location.
-pub fn compile(
-    context: &BuildContext,
+pub fn compile<'a>(
+    context: &'a BuildContext,
     python_interpreter: Option<&PythonInterpreter>,
-    bindings_crate: &BridgeModel,
+    targets: &[CompileTarget<'a>],
 ) -> Result<Vec<HashMap<String, BuildArtifact>>> {
-    let root_pkg = context.cargo_metadata.root_package().unwrap();
-    let mut targets: Vec<_> = root_pkg
-        .targets
-        .iter()
-        .filter(|target| match bindings_crate {
-            BridgeModel::Bin(_) => target.kind.contains(&"bin".to_string()),
-            _ => target.kind.contains(&"cdylib".to_string()),
-        })
-        .collect();
-    if targets.is_empty() && !bindings_crate.is_bin() {
-        // No `crate-type = ["cdylib"]` in `Cargo.toml`
-        // Let's try compile one of the target with `--crate-type cdylib`
-        let lib_target = root_pkg.targets.iter().find(|target| {
-            target
-                .kind
-                .iter()
-                .any(|k| LIB_CRATE_TYPES.contains(&k.as_str()))
-        });
-        if let Some(target) = lib_target {
-            targets.push(target);
-        }
-    }
     if context.target.is_macos() && context.universal2 {
-        compile_universal2(context, python_interpreter, bindings_crate, &targets)
+        compile_universal2(context, python_interpreter, targets)
     } else {
-        compile_targets(context, python_interpreter, bindings_crate, &targets)
+        compile_targets(context, python_interpreter, targets)
     }
 }
 
 /// Build an universal2 wheel for macos which contains both an x86 and an aarch64 binary
-fn compile_universal2(
-    context: &BuildContext,
+fn compile_universal2<'a>(
+    context: &'a BuildContext,
     python_interpreter: Option<&PythonInterpreter>,
-    bindings_crate: &BridgeModel,
-    targets: &[&cargo_metadata::Target],
+    targets: &[CompileTarget<'a>],
 ) -> Result<Vec<HashMap<String, BuildArtifact>>> {
-    let build_type = if bindings_crate.is_bin() {
-        "bin"
-    } else {
-        "cdylib"
-    };
     let mut aarch64_context = context.clone();
     aarch64_context.target = Target::from_target_triple(Some("aarch64-apple-darwin".to_string()))?;
 
-    let aarch64_artifacts = compile_targets(
-        &aarch64_context,
-        python_interpreter,
-        bindings_crate,
-        targets,
-    )
-    .context("Failed to build a aarch64 library through cargo")?;
+    let aarch64_artifacts = compile_targets(&aarch64_context, python_interpreter, targets)
+        .context("Failed to build a aarch64 library through cargo")?;
     let mut x86_64_context = context.clone();
     x86_64_context.target = Target::from_target_triple(Some("x86_64-apple-darwin".to_string()))?;
 
-    let x86_64_artifacts =
-        compile_targets(&x86_64_context, python_interpreter, bindings_crate, targets)
-            .context("Failed to build a x86_64 library through cargo")?;
+    let x86_64_artifacts = compile_targets(&x86_64_context, python_interpreter, targets)
+        .context("Failed to build a x86_64 library through cargo")?;
 
     let mut universal_artifacts = Vec::with_capacity(targets.len());
-    for (aarch64_artifact, x86_64_artifact) in aarch64_artifacts.iter().zip(x86_64_artifacts) {
+    for (bridge_model, (aarch64_artifact, x86_64_artifact)) in targets
+        .iter()
+        .map(|(_, bridge_model)| bridge_model)
+        .zip(aarch64_artifacts.iter().zip(&x86_64_artifacts))
+    {
+        let build_type = if bridge_model.is_bin() {
+            "bin"
+        } else {
+            "cdylib"
+        };
         let aarch64_artifact = aarch64_artifact.get(build_type).cloned().ok_or_else(|| {
             if build_type == "cdylib" {
                 anyhow!(
@@ -146,18 +124,17 @@ fn compile_universal2(
     Ok(universal_artifacts)
 }
 
-fn compile_targets(
-    context: &BuildContext,
+fn compile_targets<'a>(
+    context: &'a BuildContext,
     python_interpreter: Option<&PythonInterpreter>,
-    bindings_crate: &BridgeModel,
-    targets: &[&cargo_metadata::Target],
+    targets: &[CompileTarget<'a>],
 ) -> Result<Vec<HashMap<String, BuildArtifact>>> {
     let mut artifacts = Vec::with_capacity(targets.len());
-    for target in targets {
+    for (target, bridge_model) in targets {
         artifacts.push(compile_target(
             context,
             python_interpreter,
-            bindings_crate,
+            bridge_model,
             target,
         )?);
     }
@@ -167,7 +144,7 @@ fn compile_targets(
 fn compile_target(
     context: &BuildContext,
     python_interpreter: Option<&PythonInterpreter>,
-    bindings_crate: &BridgeModel,
+    bridge_model: &BridgeModel,
     binding_target: &cargo_metadata::Target,
 ) -> Result<HashMap<String, BuildArtifact>> {
     let target = &context.target;
@@ -189,6 +166,7 @@ fn compile_target(
         // `--crate-type` is stable since Rust 1.64.0
         // See https://github.com/rust-lang/cargo/pull/10838
         if target.rustc_version.semver >= RUST_1_64_0 {
+            debug!("Setting crate_type to cdylib for Rust >= 1.64.0");
             cargo_rustc.crate_type = vec!["cdylib".to_string()];
         }
     }
@@ -196,16 +174,20 @@ fn compile_target(
     let mut rust_flags = env::var_os("RUSTFLAGS");
 
     // We need to pass --bin / --lib
-    match bindings_crate {
+    match bridge_model {
         BridgeModel::Bin(..) => {
             cargo_rustc.bin.push(binding_target.name.clone());
         }
-        BridgeModel::Cffi | BridgeModel::Bindings(..) | BridgeModel::BindingsAbi3(..) => {
+        BridgeModel::Cffi
+        | BridgeModel::UniFfi
+        | BridgeModel::Bindings(..)
+        | BridgeModel::BindingsAbi3(..) => {
             cargo_rustc.lib = true;
             // https://github.com/rust-lang/rust/issues/59302#issue-422994250
             // We must only do this for libraries as it breaks binaries
             // For some reason this value is ignored when passed as rustc argument
             if context.target.is_musl_target() {
+                debug!("Setting `-C target-features=-crt-static` for musl dylib");
                 rust_flags
                     .get_or_insert_with(Default::default)
                     .push(" -C target-feature=-crt-static");
@@ -213,21 +195,19 @@ fn compile_target(
         }
     }
 
-    let module_name = &context.module_name;
-    let so_filename = match python_interpreter {
-        Some(python_interpreter) => python_interpreter.get_library_name(module_name),
-        // abi3
-        None => {
-            format!("{base}.abi3.so", base = module_name)
-        }
-    };
-
     // https://github.com/PyO3/pyo3/issues/88#issuecomment-337744403
     if target.is_macos() {
-        if let BridgeModel::Bindings(..) | BridgeModel::BindingsAbi3(..) = bindings_crate {
+        if let BridgeModel::Bindings(..) | BridgeModel::BindingsAbi3(..) = bridge_model {
             // Change LC_ID_DYLIB to the final .so name for macOS targets to avoid linking with
             // non-existent library.
             // See https://github.com/PyO3/setuptools-rust/issues/106 for detail
+            let module_name = &context.module_name;
+            let so_filename = match bridge_model {
+                BridgeModel::BindingsAbi3(..) => format!("{base}.abi3.so", base = module_name),
+                _ => python_interpreter
+                    .expect("missing python interpreter for non-abi3 wheel build")
+                    .get_library_name(module_name),
+            };
             let macos_dylib_install_name =
                 format!("link-args=-Wl,-install_name,@rpath/{}", so_filename);
             let mac_args = [
@@ -238,6 +218,7 @@ fn compile_target(
                 "-C".to_string(),
                 macos_dylib_install_name,
             ];
+            debug!("Setting additional linker args for macOS: {:?}", mac_args);
             cargo_rustc.args.extend(mac_args);
         }
     } else if target.is_emscripten() {
@@ -264,6 +245,10 @@ fn compile_target(
             emscripten_args.push("-C".to_string());
             emscripten_args.push("link-arg=-sWASM_BIGINT".to_string());
         }
+        debug!(
+            "Setting additional linker args for Emscripten: {:?}",
+            emscripten_args
+        );
         cargo_rustc.args.extend(emscripten_args);
     }
 
@@ -328,7 +313,7 @@ fn compile_target(
         build_command.env("RUSTFLAGS", flags);
     }
 
-    if let BridgeModel::BindingsAbi3(_, _) = bindings_crate {
+    if let BridgeModel::BindingsAbi3(_, _) = bridge_model {
         let is_pypy = python_interpreter
             .map(|p| p.interpreter_kind.is_pypy())
             .unwrap_or(false);
@@ -347,19 +332,24 @@ fn compile_target(
     if let Some(interpreter) = python_interpreter {
         // Target python interpreter isn't runnable when cross compiling
         if interpreter.runnable {
-            if bindings_crate.is_bindings("pyo3")
-                || bindings_crate.is_bindings("pyo3-ffi")
-                || (matches!(bindings_crate, BridgeModel::BindingsAbi3(_, _))
+            if bridge_model.is_bindings("pyo3")
+                || bridge_model.is_bindings("pyo3-ffi")
+                || (matches!(bridge_model, BridgeModel::BindingsAbi3(_, _))
                     && interpreter.interpreter_kind.is_pypy())
             {
-                build_command.env("PYO3_PYTHON", &interpreter.executable);
+                build_command
+                    .env("PYO3_PYTHON", &interpreter.executable)
+                    .env(
+                        "PYO3_ENVIRONMENT_SIGNATURE",
+                        interpreter.environment_signature(),
+                    );
             }
 
             // rust-cpython, and legacy pyo3 versions
             build_command.env("PYTHON_SYS_EXECUTABLE", &interpreter.executable);
-        } else if (bindings_crate.is_bindings("pyo3")
-            || bindings_crate.is_bindings("pyo3-ffi")
-            || (matches!(bindings_crate, BridgeModel::BindingsAbi3(_, _))
+        } else if (bridge_model.is_bindings("pyo3")
+            || bridge_model.is_bindings("pyo3-ffi")
+            || (matches!(bridge_model, BridgeModel::BindingsAbi3(_, _))
                 && interpreter.interpreter_kind.is_pypy()))
             && env::var_os("PYO3_CONFIG_FILE").is_none()
         {
@@ -386,6 +376,16 @@ fn compile_target(
         build_command.env("PYO3_CROSS_LIB_DIR", lib_dir);
     }
 
+    // Set default macOS deployment target version
+    if target.is_macos() && env::var_os("MACOSX_DEPLOYMENT_TARGET").is_none() {
+        use crate::target::rustc_macosx_target_version;
+
+        let (major, minor) = rustc_macosx_target_version(target.target_triple());
+        build_command.env("MACOSX_DEPLOYMENT_TARGET", format!("{}.{}", major, minor));
+    }
+
+    debug!("Running {:?}", build_command);
+
     let mut cargo_build = build_command
         .spawn()
         .context("Failed to run `cargo rustc`")?;
@@ -398,7 +398,9 @@ fn compile_target(
         .take()
         .expect("Cargo build should have a stdout");
     for message in cargo_metadata::Message::parse_stream(BufReader::new(stream)) {
-        match message.context("Failed to parse message coming from cargo")? {
+        let message = message.context("Failed to parse cargo metadata message")?;
+        trace!("cargo message: {:?}", message);
+        match message {
             cargo_metadata::Message::CompilerArtifact(artifact) => {
                 let package_in_metadata = context
                     .cargo_metadata
@@ -415,7 +417,7 @@ fn compile_target(
                             && !artifact.features.contains(&"rustc-dep-of-std".to_string());
                         if should_warn {
                             // This is a spurious error I don't really understand
-                            println!(
+                            eprintln!(
                                 "⚠️  Warning: The package {} wasn't listed in `cargo metadata`",
                                 package_id
                             );
@@ -487,7 +489,7 @@ fn compile_target(
 /// Currently the check is only run on linux, macOS and Windows
 pub fn warn_missing_py_init(artifact: &Path, module_name: &str) -> Result<()> {
     let py_init = format!("PyInit_{}", module_name);
-    let mut fd = File::open(&artifact)?;
+    let mut fd = File::open(artifact)?;
     let mut buffer = Vec::new();
     fd.read_to_end(&mut buffer)?;
     let mut found = false;
@@ -544,7 +546,7 @@ pub fn warn_missing_py_init(artifact: &Path, module_name: &str) -> Result<()> {
     }
 
     if !found {
-        println!(
+        eprintln!(
             "⚠️  Warning: Couldn't find the symbol `{}` in the native library. \
              Python will fail to import this module. \
              If you're using pyo3, check that `#[pymodule]` uses `{}` as module name",
