@@ -1,14 +1,20 @@
 //! The wheel format is (mostly) specified in PEP 427
 use crate::project_layout::ProjectLayout;
-use crate::{BridgeModel, Metadata21, PythonInterpreter, Target};
+use crate::target::Os;
+use crate::{
+    pyproject_toml::Format, BridgeModel, Metadata21, PyProjectToml, PythonInterpreter, Target,
+};
 use anyhow::{anyhow, bail, Context, Result};
 use flate2::write::GzEncoder;
 use flate2::Compression;
 use fs_err as fs;
 use fs_err::File;
+use ignore::overrides::Override;
 use ignore::WalkBuilder;
+use normpath::PathExt as _;
 use sha2::{Digest, Sha256};
 use std::collections::{HashMap, HashSet};
+use std::env;
 use std::ffi::OsStr;
 use std::fmt::Write as _;
 #[cfg(target_family = "unix")]
@@ -21,8 +27,10 @@ use std::path::{Path, PathBuf};
 use std::process::{Command, Output};
 use std::str;
 use tempfile::{tempdir, TempDir};
+use time::macros::datetime;
+use time::OffsetDateTime;
 use tracing::debug;
-use zip::{self, ZipWriter};
+use zip::{self, DateTime, ZipWriter};
 
 /// Allows writing the module to a wheel or add it directly to the virtualenv
 pub trait ModuleWriter {
@@ -31,6 +39,7 @@ pub trait ModuleWriter {
 
     /// Adds a file with bytes as content in target relative to the module base path
     fn add_bytes(&mut self, target: impl AsRef<Path>, bytes: &[u8]) -> Result<()> {
+        debug!("Adding {}", target.as_ref().display());
         // 0o644 is the default from the zip crate
         self.add_bytes_with_permissions(target, bytes, 0o644)
     }
@@ -46,13 +55,7 @@ pub trait ModuleWriter {
 
     /// Copies the source file to the target path relative to the module base path
     fn add_file(&mut self, target: impl AsRef<Path>, source: impl AsRef<Path>) -> Result<()> {
-        let read_failed_context = format!("Failed to read {}", source.as_ref().display());
-        let mut file = File::open(source.as_ref()).context(read_failed_context.clone())?;
-        let mut buffer = Vec::new();
-        file.read_to_end(&mut buffer).context(read_failed_context)?;
-        self.add_bytes(&target, &buffer)
-            .context(format!("Failed to write to {}", target.as_ref().display()))?;
-        Ok(())
+        self.add_file_with_permissions(target, source, 0o644)
     }
 
     /// Copies the source file the the target path relative to the module base path while setting
@@ -63,12 +66,16 @@ pub trait ModuleWriter {
         source: impl AsRef<Path>,
         permissions: u32,
     ) -> Result<()> {
-        let read_failed_context = format!("Failed to read {}", source.as_ref().display());
-        let mut file = File::open(source.as_ref()).context(read_failed_context.clone())?;
+        let target = target.as_ref();
+        let source = source.as_ref();
+        debug!("Adding {} from {}", target.display(), source.display());
+
+        let read_failed_context = format!("Failed to read {}", source.display());
+        let mut file = File::open(source).context(read_failed_context.clone())?;
         let mut buffer = Vec::new();
         file.read_to_end(&mut buffer).context(read_failed_context)?;
-        self.add_bytes_with_permissions(target.as_ref(), &buffer, permissions)
-            .context(format!("Failed to write to {}", target.as_ref().display()))?;
+        self.add_bytes_with_permissions(target, &buffer, permissions)
+            .context(format!("Failed to write to {}", target.display()))?;
         Ok(())
     }
 }
@@ -83,7 +90,7 @@ impl PathWriter {
     /// Creates a [ModuleWriter] that adds the module to the current virtualenv
     pub fn venv(target: &Target, venv_dir: &Path, bridge: &BridgeModel) -> Result<Self> {
         let interpreter =
-            PythonInterpreter::check_executable(target.get_venv_python(&venv_dir), target, bridge)?
+            PythonInterpreter::check_executable(target.get_venv_python(venv_dir), target, bridge)?
                 .ok_or_else(|| {
                     anyhow!("Expected `python` to be a python interpreter inside a virtualenv ಠ_ಠ")
                 })?;
@@ -150,7 +157,9 @@ impl PathWriter {
 
 impl ModuleWriter for PathWriter {
     fn add_directory(&mut self, path: impl AsRef<Path>) -> Result<()> {
-        fs::create_dir_all(self.base_path.join(path))?;
+        let target = self.base_path.join(path);
+        debug!("Adding directory {}", target.display());
+        fs::create_dir_all(target)?;
         Ok(())
     }
 
@@ -183,7 +192,7 @@ impl ModuleWriter for PathWriter {
         file.write_all(bytes)
             .context(format!("Failed to write to file at {}", path.display()))?;
 
-        let hash = base64::encode_config(&Sha256::digest(bytes), base64::URL_SAFE_NO_PAD);
+        let hash = base64::encode_config(Sha256::digest(bytes), base64::URL_SAFE_NO_PAD);
         self.record.push((
             target.as_ref().to_str().unwrap().to_owned(),
             hash,
@@ -200,6 +209,7 @@ pub struct WheelWriter {
     record: Vec<(String, String, usize)>,
     record_file: PathBuf,
     wheel_path: PathBuf,
+    excludes: Option<Override>,
 }
 
 impl ModuleWriter for WheelWriter {
@@ -213,8 +223,12 @@ impl ModuleWriter for WheelWriter {
         bytes: &[u8],
         permissions: u32,
     ) -> Result<()> {
+        let target = target.as_ref();
+        if self.exclude(target) {
+            return Ok(());
+        }
         // The zip standard mandates using unix style paths
-        let target = target.as_ref().to_str().unwrap().replace('\\', "/");
+        let target = target.to_str().unwrap().replace('\\', "/");
 
         // Unlike users which can use the develop subcommand, the tests have to go through
         // packing a zip which pip than has to unpack. This makes this 2-3 times faster
@@ -223,13 +237,19 @@ impl ModuleWriter for WheelWriter {
         } else {
             zip::CompressionMethod::Deflated
         };
-        let options = zip::write::FileOptions::default()
+
+        let mut options = zip::write::FileOptions::default()
             .unix_permissions(permissions)
             .compression_method(compression_method);
+        let mtime = self.mtime().ok();
+        if let Some(mtime) = mtime {
+            options = options.last_modified_time(mtime);
+        }
+
         self.zip.start_file(target.clone(), options)?;
         self.zip.write_all(bytes)?;
 
-        let hash = base64::encode_config(&Sha256::digest(bytes), base64::URL_SAFE_NO_PAD);
+        let hash = base64::encode_config(Sha256::digest(bytes), base64::URL_SAFE_NO_PAD);
         self.record.push((target, hash, bytes.len()));
 
         Ok(())
@@ -245,6 +265,7 @@ impl WheelWriter {
         wheel_dir: &Path,
         metadata21: &Metadata21,
         tags: &[String],
+        excludes: Option<Override>,
     ) -> Result<WheelWriter> {
         let wheel_path = wheel_dir.join(format!(
             "{}-{}-{}.whl",
@@ -260,6 +281,7 @@ impl WheelWriter {
             record: Vec::new(),
             record_file: metadata21.get_dist_info_dir().join("RECORD"),
             wheel_path,
+            excludes,
         };
 
         write_dist_info(&mut builder, metadata21, tags)?;
@@ -274,10 +296,12 @@ impl WheelWriter {
         metadata21: &Metadata21,
     ) -> Result<()> {
         if let Some(python_module) = &project_layout.python_module {
-            let absolute_path = fs::canonicalize(python_module)?;
+            let absolute_path = python_module.normalize()?.into_path_buf();
             if let Some(python_path) = absolute_path.parent().and_then(|p| p.to_str()) {
                 let name = metadata21.get_distribution_escaped();
-                self.add_bytes(format!("{}.pth", name), python_path.as_bytes())?;
+                let target = format!("{}.pth", name);
+                debug!("Adding {} from {}", target, python_path);
+                self.add_bytes(target, python_path.as_bytes())?;
             } else {
                 println!("⚠️ source code path contains non-Unicode sequences, editable installs may not work.");
             }
@@ -285,6 +309,27 @@ impl WheelWriter {
         Ok(())
     }
 
+    /// Returns `true` if the given path should be excluded
+    fn exclude(&self, path: impl AsRef<Path>) -> bool {
+        if let Some(excludes) = &self.excludes {
+            excludes.matched(path.as_ref(), false).is_whitelist()
+        } else {
+            false
+        }
+    }
+
+    /// Returns a DateTime representing the value SOURCE_DATE_EPOCH environment variable
+    /// Note that the earliest timestamp a zip file can represent is 1980-01-01
+    fn mtime(&self) -> Result<DateTime> {
+        let epoch: i64 = env::var("SOURCE_DATE_EPOCH")?.parse()?;
+        let dt = OffsetDateTime::from_unix_timestamp(epoch)?;
+        let min_dt = datetime!(1980-01-01 0:00 UTC);
+        let dt = dt.max(min_dt);
+
+        let dt = DateTime::from_time(dt).map_err(|_| anyhow!("Failed to build zip DateTime"))?;
+        Ok(dt)
+    }
+
     /// Creates the record file and finishes the zip
     pub fn finish(mut self) -> Result<PathBuf, io::Error> {
         let compression_method = if cfg!(feature = "faster-tests") {
@@ -292,8 +337,15 @@ impl WheelWriter {
         } else {
             zip::CompressionMethod::Deflated
         };
-        let options = zip::write::FileOptions::default().compression_method(compression_method);
+
+        let mut options = zip::write::FileOptions::default().compression_method(compression_method);
+        let mtime = self.mtime().ok();
+        if let Some(mtime) = mtime {
+            options = options.last_modified_time(mtime);
+        }
+
         let record_filename = self.record_file.to_str().unwrap().replace('\\', "/");
+        debug!("Adding {}", record_filename);
         self.zip.start_file(&record_filename, options)?;
         for (filename, hash, len) in self.record {
             self.zip
@@ -313,6 +365,7 @@ pub struct SDistWriter {
     tar: tar::Builder<GzEncoder<File>>,
     path: PathBuf,
     files: HashSet<PathBuf>,
+    excludes: Option<Override>,
 }
 
 impl ModuleWriter for SDistWriter {
@@ -327,10 +380,15 @@ impl ModuleWriter for SDistWriter {
         permissions: u32,
     ) -> Result<()> {
         let target = target.as_ref();
+        if self.exclude(target) {
+            return Ok(());
+        }
+
         if self.files.contains(target) {
             // Ignore duplicate files
             return Ok(());
         }
+
         let mut header = tar::Header::new_gnu();
         header.set_size(bytes.len() as u64);
         header.set_mode(permissions);
@@ -348,17 +406,22 @@ impl ModuleWriter for SDistWriter {
 
     fn add_file(&mut self, target: impl AsRef<Path>, source: impl AsRef<Path>) -> Result<()> {
         let source = source.as_ref();
+        if self.exclude(source) {
+            return Ok(());
+        }
         let target = target.as_ref();
         if source == self.path {
-            bail!(
-            "Attempting to include the sdist output tarball {} into itself! Check 'cargo package --list' output.",
-            source.display()
+            eprintln!(
+                "⚠️  Warning: Attempting to include the sdist output tarball {} into itself! Check 'cargo package --list' output.",
+                source.display()
             );
+            return Ok(());
         }
         if self.files.contains(target) {
             // Ignore duplicate files
             return Ok(());
         }
+        debug!("Adding {} from {}", target.display(), source.display());
 
         self.tar
             .append_path_with_name(source, target)
@@ -374,7 +437,11 @@ impl ModuleWriter for SDistWriter {
 
 impl SDistWriter {
     /// Create a source distribution .tar.gz which can be subsequently expanded
-    pub fn new(wheel_dir: impl AsRef<Path>, metadata21: &Metadata21) -> Result<Self, io::Error> {
+    pub fn new(
+        wheel_dir: impl AsRef<Path>,
+        metadata21: &Metadata21,
+        excludes: Option<Override>,
+    ) -> Result<Self, io::Error> {
         let path = wheel_dir.as_ref().join(format!(
             "{}-{}.tar.gz",
             &metadata21.get_distribution_escaped(),
@@ -389,9 +456,19 @@ impl SDistWriter {
             tar,
             path,
             files: HashSet::new(),
+            excludes,
         })
     }
 
+    /// Returns `true` if the given path should be excluded
+    fn exclude(&self, path: impl AsRef<Path>) -> bool {
+        if let Some(excludes) = &self.excludes {
+            excludes.matched(path.as_ref(), false).is_whitelist()
+        } else {
+            false
+        }
+    }
+
     /// Finished the .tar.gz archive
     pub fn finish(mut self) -> Result<PathBuf, io::Error> {
         self.tar.finish()?;
@@ -446,7 +523,7 @@ where
     I: IntoIterator<Item = S>,
     S: AsRef<OsStr>,
 {
-    Command::new(&python)
+    Command::new(python)
         .args(args)
         .output()
         .context(format!("Failed to run python at {:?}", &python))
@@ -472,7 +549,7 @@ fn cffi_header(crate_dir: &Path, target_dir: &Path, tempdir: &TempDir) -> Result
             );
         }
 
-        let mut config = cbindgen::Config::from_root_or_default(&crate_dir);
+        let mut config = cbindgen::Config::from_root_or_default(crate_dir);
         config.defines = HashMap::new();
         config.include_guard = None;
 
@@ -486,6 +563,7 @@ fn cffi_header(crate_dir: &Path, target_dir: &Path, tempdir: &TempDir) -> Result
 
         let header = tempdir.as_ref().join("header.h");
         bindings.write_to_file(&header);
+        debug!("Generated header.h at {}", header.display());
         Ok(header)
     }
 }
@@ -526,7 +604,7 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
         header = header.display(),
     );
 
-    let output = call_python(python, &["-c", &cffi_invocation])?;
+    let output = call_python(python, ["-c", &cffi_invocation])?;
     let install_cffi = if !output.status.success() {
         // First, check whether the error was cffi not being installed
         let last_line = str::from_utf8(&output.stderr)?.lines().last().unwrap_or("");
@@ -536,7 +614,7 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
             // https://stackoverflow.com/a/42580137/3549270
             let output = call_python(
                 python,
-                &["-c", "import sys\nprint(sys.base_prefix != sys.prefix)"],
+                ["-c", "import sys\nprint(sys.base_prefix != sys.prefix)"],
             )?;
 
             match str::from_utf8(&output.stdout)?.trim() {
@@ -565,7 +643,16 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
     println!("⚠️ cffi not found. Trying to install it");
     // Call pip through python to don't do the wrong thing when python and pip
     // are coming from different environments
-    let output = call_python(python, &["-m", "pip", "install", "cffi"])?;
+    let output = call_python(
+        python,
+        [
+            "-m",
+            "pip",
+            "install",
+            "--disable-pip-version-check",
+            "cffi",
+        ],
+    )?;
     if !output.status.success() {
         bail!(
             "Installing cffi with `{:?} -m pip install cffi` failed: {}\n--- Stdout:\n{}\n--- Stderr:\n{}\n---\nPlease install cffi yourself.",
@@ -578,7 +665,7 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
     println!("🎁 Installed cffi");
 
     // Try again
-    let output = call_python(python, &["-c", &cffi_invocation])?;
+    let output = call_python(python, ["-c", &cffi_invocation])?;
     handle_cffi_call_result(python, tempdir, &ffi_py, &output)
 }
 
@@ -601,7 +688,7 @@ fn handle_cffi_call_result(
         // Don't swallow warnings
         io::stderr().write_all(&output.stderr)?;
 
-        let ffi_py_content = fs::read_to_string(&ffi_py)?;
+        let ffi_py_content = fs::read_to_string(ffi_py)?;
         tempdir.close()?;
         Ok(ffi_py_content)
     }
@@ -617,6 +704,7 @@ pub fn write_bindings_module(
     python_interpreter: Option<&PythonInterpreter>,
     target: &Target,
     editable: bool,
+    pyproject_toml: Option<&PyProjectToml>,
 ) -> Result<()> {
     let ext_name = &project_layout.extension_name;
     let so_filename = match python_interpreter {
@@ -633,28 +721,28 @@ pub fn write_bindings_module(
     };
 
     if let Some(python_module) = &project_layout.python_module {
-        if !editable {
-            write_python_part(writer, python_module)
-                .context("Failed to add the python module to the package")?;
-        }
-
         if editable {
             let target = project_layout.rust_module.join(&so_filename);
             // Remove existing so file to avoid triggering SIGSEV in running process
             // See https://github.com/PyO3/maturin/issues/758
+            debug!("Removing {}", target.display());
             let _ = fs::remove_file(&target);
-            fs::copy(&artifact, &target).context(format!(
+
+            debug!("Copying {} to {}", artifact.display(), target.display());
+            fs::copy(artifact, &target).context(format!(
                 "Failed to copy {} to {}",
                 artifact.display(),
                 target.display()
             ))?;
-        }
+        } else {
+            write_python_part(writer, python_module, pyproject_toml)
+                .context("Failed to add the python module to the package")?;
 
-        if !editable {
             let relative = project_layout
                 .rust_module
-                .strip_prefix(python_module.parent().unwrap())?;
-            writer.add_file_with_permissions(relative.join(&so_filename), &artifact, 0o755)?;
+                .strip_prefix(python_module.parent().unwrap())
+                .unwrap();
+            writer.add_file_with_permissions(relative.join(&so_filename), artifact, 0o755)?;
         }
     } else {
         let module = PathBuf::from(module_name);
@@ -676,14 +764,11 @@ if hasattr({module_name}, "__all__"):
             .rust_module
             .join(format!("{}.pyi", module_name));
         if type_stub.exists() {
-            writer.add_bytes(
-                &module.join("__init__.pyi"),
-                &fs_err::read(type_stub).context("Failed to read type stub file")?,
-            )?;
-            writer.add_bytes(&module.join("py.typed"), b"")?;
             println!("📖 Found type stub file at {}.pyi", module_name);
+            writer.add_file(&module.join("__init__.pyi"), type_stub)?;
+            writer.add_bytes(&module.join("py.typed"), b"")?;
         }
-        writer.add_file_with_permissions(&module.join(so_filename), &artifact, 0o755)?;
+        writer.add_file_with_permissions(&module.join(so_filename), artifact, 0o755)?;
     }
 
     Ok(())
@@ -700,6 +785,7 @@ pub fn write_cffi_module(
     artifact: &Path,
     python: &Path,
     editable: bool,
+    pyproject_toml: Option<&PyProjectToml>,
 ) -> Result<()> {
     let cffi_declarations = generate_cffi_declarations(crate_dir, target_dir, python)?;
 
@@ -707,15 +793,15 @@ pub fn write_cffi_module(
 
     if let Some(python_module) = &project_layout.python_module {
         if !editable {
-            write_python_part(writer, python_module)
+            write_python_part(writer, python_module, pyproject_toml)
                 .context("Failed to add the python module to the package")?;
         }
 
         if editable {
-            let base_path = python_module.join(&module_name);
+            let base_path = python_module.join(module_name);
             fs::create_dir_all(&base_path)?;
             let target = base_path.join("native.so");
-            fs::copy(&artifact, &target).context(format!(
+            fs::copy(artifact, &target).context(format!(
                 "Failed to copy {} to {}",
                 artifact.display(),
                 target.display()
@@ -726,7 +812,8 @@ pub fn write_cffi_module(
 
         let relative = project_layout
             .rust_module
-            .strip_prefix(python_module.parent().unwrap())?;
+            .strip_prefix(python_module.parent().unwrap())
+            .unwrap();
         module = relative.join(&project_layout.extension_name);
         if !editable {
             writer.add_directory(&module)?;
@@ -738,19 +825,193 @@ pub fn write_cffi_module(
             .rust_module
             .join(format!("{}.pyi", module_name));
         if type_stub.exists() {
-            writer.add_bytes(
-                &module.join("__init__.pyi"),
-                &fs_err::read(type_stub).context("Failed to read type stub file")?,
-            )?;
-            writer.add_bytes(&module.join("py.typed"), b"")?;
             println!("📖 Found type stub file at {}.pyi", module_name);
+            writer.add_file(&module.join("__init__.pyi"), type_stub)?;
+            writer.add_bytes(&module.join("py.typed"), b"")?;
         }
     };
 
     if !editable || project_layout.python_module.is_none() {
         writer.add_bytes(&module.join("__init__.py"), cffi_init_file().as_bytes())?;
         writer.add_bytes(&module.join("ffi.py"), cffi_declarations.as_bytes())?;
-        writer.add_file_with_permissions(&module.join("native.so"), &artifact, 0o755)?;
+        writer.add_file_with_permissions(&module.join("native.so"), artifact, 0o755)?;
+    }
+
+    Ok(())
+}
+
+/// uniffi.toml
+#[derive(Debug, serde::Deserialize)]
+struct UniFfiToml {
+    #[serde(default)]
+    bindings: HashMap<String, UniFfiBindingsConfig>,
+}
+
+/// `bindings` section of uniffi.toml
+#[derive(Debug, serde::Deserialize)]
+struct UniFfiBindingsConfig {
+    cdylib_name: Option<String>,
+}
+
+#[derive(Debug, Clone)]
+struct UniFfiBindings {
+    name: String,
+    cdylib: String,
+    path: PathBuf,
+}
+
+fn generate_uniffi_bindings(
+    crate_dir: &Path,
+    target_dir: &Path,
+    target_os: Os,
+) -> Result<UniFfiBindings> {
+    let binding_dir = target_dir.join("maturin").join("uniffi");
+    fs::create_dir_all(&binding_dir)?;
+
+    let pattern = crate_dir.join("src").join("*.udl");
+    let udls = glob::glob(pattern.to_str().unwrap())?
+        .map(|p| p.unwrap())
+        .collect::<Vec<_>>();
+    if udls.is_empty() {
+        bail!("No UDL files found in {}", crate_dir.join("src").display());
+    } else if udls.len() > 1 {
+        bail!(
+            "Multiple UDL files found in {}",
+            crate_dir.join("src").display()
+        );
+    }
+
+    let udl = &udls[0];
+    let config_file = crate_dir.join("uniffi.toml");
+    let mut cdylib_name = None;
+    let config_file = {
+        if config_file.is_file() {
+            let uniffi_toml: UniFfiToml =
+                toml_edit::easy::from_str(&fs::read_to_string(&config_file)?)?;
+            cdylib_name = uniffi_toml
+                .bindings
+                .get("python")
+                .and_then(|py| py.cdylib_name.clone());
+            Some(
+                config_file
+                    .as_path()
+                    .try_into()
+                    .expect("path contains non-utf8"),
+            )
+        } else {
+            None
+        }
+    };
+    uniffi_bindgen::generate_bindings(
+        udl.as_path().try_into().expect("path contains non-utf8"),
+        config_file,
+        vec!["python"],
+        Some(
+            binding_dir
+                .as_path()
+                .try_into()
+                .expect("path contains non-utf8"),
+        ),
+        None,
+        false,
+    )?;
+    let py_binding_name = udl.file_stem().unwrap();
+    let py_binding = binding_dir.join(py_binding_name).with_extension("py");
+    let name = py_binding_name.to_str().unwrap().to_string();
+
+    // uniffi bindings hardcoded the extension filenames
+    let cdylib_name = match cdylib_name {
+        Some(name) => name,
+        None => format!("uniffi_{}", name),
+    };
+    let cdylib = match target_os {
+        Os::Macos => format!("lib{}.dylib", cdylib_name),
+        Os::Windows => format!("{}.dll", cdylib_name),
+        _ => format!("lib{}.so", cdylib_name),
+    };
+
+    Ok(UniFfiBindings {
+        name,
+        cdylib,
+        path: py_binding,
+    })
+}
+
+/// Creates the uniffi module with the shared library
+#[allow(clippy::too_many_arguments)]
+pub fn write_uniffi_module(
+    writer: &mut impl ModuleWriter,
+    project_layout: &ProjectLayout,
+    crate_dir: &Path,
+    target_dir: &Path,
+    module_name: &str,
+    artifact: &Path,
+    target_os: Os,
+    editable: bool,
+    pyproject_toml: Option<&PyProjectToml>,
+) -> Result<()> {
+    let UniFfiBindings {
+        name: binding_name,
+        cdylib,
+        path: uniffi_binding,
+    } = generate_uniffi_bindings(crate_dir, target_dir, target_os)?;
+    let py_init = format!("from .{} import *  # NOQA\n", binding_name);
+
+    let module;
+
+    if let Some(python_module) = &project_layout.python_module {
+        if !editable {
+            write_python_part(writer, python_module, pyproject_toml)
+                .context("Failed to add the python module to the package")?;
+        }
+
+        if editable {
+            let base_path = python_module.join(module_name);
+            fs::create_dir_all(&base_path)?;
+            let target = base_path.join(&cdylib);
+            fs::copy(artifact, &target).context(format!(
+                "Failed to copy {} to {}",
+                artifact.display(),
+                target.display()
+            ))?;
+
+            File::create(base_path.join("__init__.py"))?.write_all(py_init.as_bytes())?;
+            let target = base_path.join(&binding_name).with_extension("py");
+            fs::copy(&uniffi_binding, &target).context(format!(
+                "Failed to copy {} to {}",
+                uniffi_binding.display(),
+                target.display()
+            ))?;
+        }
+
+        let relative = project_layout
+            .rust_module
+            .strip_prefix(python_module.parent().unwrap())
+            .unwrap();
+        module = relative.join(&project_layout.extension_name);
+        if !editable {
+            writer.add_directory(&module)?;
+        }
+    } else {
+        module = PathBuf::from(module_name);
+        writer.add_directory(&module)?;
+        let type_stub = project_layout
+            .rust_module
+            .join(format!("{}.pyi", module_name));
+        if type_stub.exists() {
+            println!("📖 Found type stub file at {}.pyi", module_name);
+            writer.add_file(&module.join("__init__.pyi"), type_stub)?;
+            writer.add_bytes(&module.join("py.typed"), b"")?;
+        }
+    };
+
+    if !editable || project_layout.python_module.is_none() {
+        writer.add_bytes(&module.join("__init__.py"), py_init.as_bytes())?;
+        writer.add_file(
+            module.join(binding_name).with_extension("py"),
+            uniffi_binding,
+        )?;
+        writer.add_file_with_permissions(&module.join(cdylib), artifact, 0o755)?;
     }
 
     Ok(())
@@ -761,7 +1022,7 @@ pub fn write_bin(
     writer: &mut impl ModuleWriter,
     artifact: &Path,
     metadata: &Metadata21,
-    bin_name: &OsStr,
+    bin_name: &str,
 ) -> Result<()> {
     let data_dir = PathBuf::from(format!(
         "{}-{}.data",
@@ -777,20 +1038,80 @@ pub fn write_bin(
     Ok(())
 }
 
+/// Adds a wrapper script that start the wasm binary through wasmtime.
+///
+/// Note that the wasm binary needs to be written separately by [write_bin]
+pub fn write_wasm_launcher(
+    writer: &mut impl ModuleWriter,
+    metadata: &Metadata21,
+    bin_name: &str,
+) -> Result<()> {
+    let entrypoint_script = format!(
+        r#"from pathlib import Path
+
+from wasmtime import Store, Module, Engine, WasiConfig, Linker
+
+import sysconfig
+
+def main():
+    # The actual executable
+    program_location = Path(sysconfig.get_path("scripts")).joinpath("{}")
+    # wasmtime-py boilerplate
+    engine = Engine()
+    store = Store(engine)
+    # TODO: is there an option to just get the default of the wasmtime cli here?
+    wasi = WasiConfig()
+    wasi.inherit_argv()
+    wasi.inherit_env()
+    wasi.inherit_stdout()
+    wasi.inherit_stderr()
+    wasi.inherit_stdin()
+    # TODO: Find a real solution here. Maybe there's an always allow callback?
+    # Even fancier would be something configurable in pyproject.toml
+    wasi.preopen_dir(".", ".")
+    store.set_wasi(wasi)
+    linker = Linker(engine)
+    linker.define_wasi()
+    module = Module.from_file(store.engine, str(program_location))
+    linking1 = linker.instantiate(store, module)
+    # TODO: this is taken from https://docs.wasmtime.dev/api/wasmtime/struct.Linker.html#method.get_default
+    #       is this always correct?
+    start = linking1.exports(store).get("") or linking1.exports(store)["_start"]
+    start(store)
+
+if __name__ == '__main__':
+    main()
+    "#,
+        bin_name
+    );
+
+    // We can't use add_file since we want to mark the file as executable
+    let launcher_path = Path::new(&metadata.get_distribution_escaped())
+        .join(bin_name.replace('-', "_"))
+        .with_extension("py");
+    writer.add_bytes_with_permissions(&launcher_path, entrypoint_script.as_bytes(), 0o755)?;
+    Ok(())
+}
+
 /// Adds the python part of a mixed project to the writer,
 pub fn write_python_part(
     writer: &mut impl ModuleWriter,
     python_module: impl AsRef<Path>,
+    pyproject_toml: Option<&PyProjectToml>,
 ) -> Result<()> {
-    for absolute in WalkBuilder::new(&python_module).hidden(false).build() {
+    let python_module = python_module.as_ref();
+    for absolute in WalkBuilder::new(python_module).hidden(false).build() {
         let absolute = absolute?.into_path();
-        let relative = absolute.strip_prefix(python_module.as_ref().parent().unwrap())?;
+        let relative = absolute
+            .strip_prefix(python_module.parent().unwrap())
+            .unwrap();
         if absolute.is_dir() {
             writer.add_directory(relative)?;
         } else {
             // Ignore native libraries from develop, if any
             if let Some(extension) = relative.extension() {
                 if extension.to_string_lossy() == "so" {
+                    debug!("Ignoring native library {}", relative.display());
                     continue;
                 }
             }
@@ -800,6 +1121,30 @@ pub fn write_python_part(
         }
     }
 
+    // Include additional files
+    if let Some(pyproject) = pyproject_toml {
+        let pyproject_dir = python_module.parent().unwrap();
+        if let Some(glob_patterns) = pyproject.include() {
+            for pattern in glob_patterns
+                .iter()
+                .filter_map(|glob_pattern| glob_pattern.targets(Format::Sdist))
+            {
+                println!("📦 Including files matching \"{}\"", pattern);
+                for source in glob::glob(&pyproject_dir.join(pattern).to_string_lossy())
+                    .expect("No files found for pattern")
+                    .filter_map(Result::ok)
+                {
+                    let target = source.strip_prefix(pyproject_dir)?.to_path_buf();
+                    if source.is_dir() {
+                        writer.add_directory(target)?;
+                    } else {
+                        writer.add_file(target, source)?;
+                    }
+                }
+            }
+        }
+    }
+
     Ok(())
 }
 
@@ -880,7 +1225,7 @@ pub fn add_data(writer: &mut impl ModuleWriter, data: Option<&Path>) -> Result<(
                     .build()
                 {
                     let file = file?;
-                    let relative = file.path().strip_prefix(data.parent().unwrap())?;
+                    let relative = file.path().strip_prefix(data.parent().unwrap()).unwrap();
 
                     if file.path_is_symlink() {
                         // Copy the actual file contents, not the link, so that you can create a
@@ -902,3 +1247,44 @@ pub fn add_data(writer: &mut impl ModuleWriter, data: Option<&Path>) -> Result<(
     }
     Ok(())
 }
+
+#[cfg(test)]
+mod tests {
+    use ignore::overrides::OverrideBuilder;
+
+    use super::*;
+
+    #[test]
+    // The mechanism is the same for wheel_writer
+    fn sdist_writer_excludes() -> Result<(), Box<dyn std::error::Error>> {
+        let metadata = Metadata21::default();
+        let perm = 0o777;
+
+        // No excludes
+        let tmp_dir = TempDir::new()?;
+        let mut writer = SDistWriter::new(&tmp_dir, &metadata, None)?;
+        assert!(writer.files.is_empty());
+        writer.add_bytes_with_permissions("test", &[], perm)?;
+        assert_eq!(writer.files.len(), 1);
+        writer.finish()?;
+        tmp_dir.close()?;
+
+        // A test filter
+        let tmp_dir = TempDir::new()?;
+        let mut excludes = OverrideBuilder::new(&tmp_dir);
+        excludes.add("test*")?;
+        excludes.add("!test2")?;
+        let mut writer = SDistWriter::new(&tmp_dir, &metadata, Some(excludes.build()?))?;
+        writer.add_bytes_with_permissions("test1", &[], perm)?;
+        writer.add_bytes_with_permissions("test3", &[], perm)?;
+        assert!(writer.files.is_empty());
+        writer.add_bytes_with_permissions("test2", &[], perm)?;
+        assert!(!writer.files.is_empty());
+        writer.add_bytes_with_permissions("yes", &[], perm)?;
+        assert_eq!(writer.files.len(), 2);
+        writer.finish()?;
+        tmp_dir.close()?;
+
+        Ok(())
+    }
+}
